/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Header: /usr/local/db4/cvsroot/steve/lbback/src/builtin2.C,v 1.3 1995/02/13 17:49:15 steve Exp $
#ifndef BUILTIN2_H
#include "builtin2.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

static Value::tv_type_id one_undefined[] = { Value::UNDEFINED };

BUILTIN_FIRST_DEFINITION(length_pred,1,1,one_undefined,1,1)
{
  if (m || n); // Suppress warning.
  V_Collection *list;
  if (in_tv[0]->tv_type() == Value::LIST) list = in_tv[0]->down_to_LIST();
  else if (in_tv[0]->tv_type() == Value::SLIST) list = in_tv[0]->down_to_SLIST();
  // We will eventually also have a length predicate for strings
  // and DNA sequences.
  else {
    ql_eval_error(t, "Argument 1 must be a LIST or SLIST in");
    return 0;
  }
  out_tv[0] = Int::make(list->cardinality());
  return (QL_Iterator_State*) 1;
}


static Value::tv_type_id one_set[] = { Value::SET };

BUILTIN_FIRST_DEFINITION(cardinality_pred,1,1,one_set,1,1)
{
  out_tv[0] = Int::make(in_tv[0]->down_to_SET()->cardinality());
  if (m || n || t);  // Supress warning;
  return (QL_Iterator_State*) 1;
}


BUILTIN_FIRST_DEFINITION(commit_pred,0,0,0,0,1)
{
  m = n = (out_tv == in_tv);  t = 0; // Suppress warnings.
  QL_DD::commit();
  return (QL_Iterator_State*)1;
}  

static Value::tv_type_id all_steps_types[]
  = { Value::MATERIAL_POINTER };

BUILTIN_FIRST_DEFINITION(all_steps_bf_pred_first,1,1,all_steps_types,1,1)
{
  Cursor_Step *c = in_tv[0]->down_to_MATERIAL_POINTER()->step_cursor();
  if (!c) return 0;
  Step *s = c->first();
  if (!s) { delete c; return 0; }
  out_tv[0] = s;
  t = 0; m = n = 0; // Suppress warnings.
  return (QL_Iterator_State*) c;
}


BUILTIN_NEXT_DEFINITION(all_steps_bf_pred_next,1)
{
  ABORT(n != 1, "Wrong number of free arguments");
  Cursor_Step *c = (Cursor_Step*) s;
  Step *step = c->next();
  if (!step) { delete c; return 0; }
  out_tv[0] = step;
  t = 0;
  return s;
}


BUILTIN_FIRST_DEFINITION(element_pred_first,-1,-1,one_undefined,1,0)
{
  // STATE: done, regression tested, memory tested
  // BY: amarquis
  // QL_Term*t -- for ql_eval_error
  // int m -- number of bound arguments
  // int n -- number of free arguments
  // const Value in_tv[m] -- bound arguments
  // Value out_tv[m] -- output_bindings
  // Return value: a cursor point to the first element of the first bound term

  if (m < 1 ||
      (in_tv[0]->tv_type() != Value::LIST
       && in_tv[0]->tv_type() != Value::SET
       && in_tv[0]->tv_type() != Value::SLIST)) {
      ql_eval_error(t, "First argument must be a LIST, SLIST, or SET for");
      return 0;
    }

  if (m + n < 2) {
    ql_eval_error(t, "Require at least two arguments for");
    return 0;
  }

  V_Collection *p = (V_Collection*) in_tv[0];
  if (p->cardinality() < 1)
    {
      // Empty collection -- don't iterate over it.
      return 0;
    }

  // The following is temporary special case code, 
  // to be used until there are general underbound
  // methods.
  if (n == 0) {
    if (m != 2) {
      ql_eval_error(t, "If more than one argument is bound, ",
		    "adorment must be [bb] for");
      return 0;
    }
    V_Cursor *c = p->cursor_of();
    for (Value *z = c->first(); z; z = c->next()) {
      clear_error();
      int found = (*z == *in_tv[1]);
      if (errmsg()) { ql_eval_error(t, errmsg(), " in"); delete c; return 0; }
      if (found) return (QL_Iterator_State*) c;
    }
    delete c; return 0;
  }

  if (n == 1)
    {
      // Bind contents of the collection in turn to the one free variable
      V_Cursor *c = p->cursor_of();

      Value *v = c->first();
      if (!v)
	{
	  delete c;
	  return 0;
	}
      out_tv[0] = v;

      // Suppress warnings.
      t = 0;
      m = n = 0;

      return (QL_Iterator_State*) c;
    }
  else
    {
      // Bind the contents of the first collection to all the free variables...

      // Check that each element of the collection is the right type and size
      V_Cursor *c = p->cursor_of();
      for (Value *v = c->first(); v; v = c->next()) {
	if (((V_Collection*)v)->cardinality() != n) {
	  ql_eval_error(t, "The number of items in a sub-list of the ",
			"first term does not match the number of free ",
			"variables in");
	  delete c;
	  return 0;
	}
      }
      V_Collection* lv = (V_Collection*) c->first();
      if (!lv) {
	delete c;
	return 0;
      }
      V_Cursor *clv = lv->cursor_of();
      Value *tv;
      int i;
      for (tv = clv->first(), i = 0; tv; tv = clv->next(), i++) out_tv[i] = tv;
      delete clv;
      return (QL_Iterator_State*) c;
    }
}


QL_NEXT_HEADER(element_pred_next)
{
  if (m || t || n) ; // suppress warning

  V_Cursor *c = (V_Cursor*)s;
  if (n == 0) { // Adornment [bb] -- a special case.
    for (Value* z = c->next() ; z ; z = c->next()) {
      clear_error();
      int found = (*z == *in[1]);
      if (errmsg()) { ql_eval_error(t, errmsg(), " in"); delete c; return 0; }
      if (found) return (QL_Iterator_State*) c;
    }
    delete c; return 0;
  }

  if (n == 1) {
    // Bind contents of the collection in turn to the one free variable
    // State is the cursor

    Value *v = c->next();
    if (!v) {
      // Done iterating over this collection
      delete c;
      return 0;
    }
    out[0] = v;
    return (QL_Iterator_State*) c;
  }
  else {
    // Bind the contents of the first collection to all the free variables...
    // It is important that the collection sizes matching the number of free
    // variables was checked before entering this part of the code.

    V_Collection* lv = (V_Collection*) c->next();
    if (!lv) {
      // Done iterating through sub-collections.
      delete c;
      return 0;
    }
    V_Cursor *clv = lv->cursor_of();
    Value *tv;
    int i;
    for (tv = clv->first(), i = 0; tv; tv = clv->next(), i++) out[i] = tv;
    delete clv;
    return (QL_Iterator_State*) c;
  }
}


class ith_pred_state
{
public:
  ith_pred_state(V_Collection *list)
    : _index(0), _cursor(list->cursor_of()) {}

  ~ith_pred_state() { delete cursor(); }
  Value *get_next() { _index++; return cursor()->next(); }
  SLOT(int,index);
  SLOT(V_Cursor*,cursor);
};


BUILTIN_FIRST_DEFINITION(ith_pred_first,-1,-1,one_undefined,1,0)
{
  // STATE: not done
  // BY: amarquis
  // QL_Term*t -- for ql_eval_error
  // int m -- number of bound arguments
  // int n -- number of free arguments
  // const Value in_tv[m] -- bound arguments
  // Value out_tv[m] -- output_bindings
  // Return value: a ith_pred_state object, containing a cursor and
  // the next index.

  if (m < 1 || m > 2 || (m + n) != 3) {
    ql_eval_error(t, "Only adornments [bff] and [bbf] are allowed for");
    return 0;
  }

  V_Collection *list;
  if (in_tv[0]->tv_type() == Value::LIST)  list = in_tv[0]->down_to_LIST();
  else if (in_tv[0]->tv_type() == Value::SLIST)
    list = in_tv[0]->down_to_SLIST();
  else {
    ql_eval_error(t, "First argument must be a LIST or SLIST for");
    return 0;
  }

  if (2 == m) {
    int index;
    if (!coerce_to_int(t, "Argument 2 ", in_tv[1], &index)) return 0;
    if (index < 0 || index >= list->cardinality()) {
      ql_eval_error(t, "The second argument is out of bounds in");
      return 0;
    }
    out_tv[0] = list->get(index);
    return (QL_Iterator_State*)1;
  }
    
  if (list->cardinality() < 1) return 0;  // Empty collection -- don't iterate over it.

  ith_pred_state *state = new ith_pred_state(list);
  if (!state->cursor()) {
    ql_eval_error(t, "Could not produce an ObjectStore cursor while evaluating");
    return 0;
  }

  Value *v = state->cursor()->first();

  if (!v) { delete state;  return 0; }

  if (n == 2) {
    // Bind indexes to the first free variable and 
    // the corresponding list element to the other free variable.
    out_tv[0] = Int::make(state->index());
    out_tv[1] = v;
    return (QL_Iterator_State*) state;
  }
  else ABORT(1, "Not implemented yet");
  return 0;
}


BUILTIN_NEXT_DEFINITION(ith_pred_next,0)
{
  // STATE: not done
  // BY: amarquis
  // QL_Term*t -- for ql_eval_error
  // int n -- number of free arguments
  // const Value in_tv[m] -- bound arguments
  // Value out_tv[m] -- output_bindings
  // QL_Iterator_State* s -- state of last call
  // Return value -- a cursor pointing to the ith_element of the first bound
  // term used for this set of bindings. 

  if (1 == (int)s) return 0; // This was the version in which the index
			     // was bound.

  if (n == 2) {
    // Bind contents of the collection in turn to the one free variable
    // State is the cursor
    ith_pred_state *state = (ith_pred_state*)s;
    Value *v = state->get_next();
    if (!v){
      // Done iterating over this collection
      delete state;  return 0;
    }
    out_tv[0] = Int::make(state->index());
    out_tv[1] = v;
    return s;
  }
  else  ABORT(1,"Not implemented yet");
  if (t);
  return 0;
}


BUILTIN_FIRST_DEFINITION(make_list_pred,-1,1,one_undefined,1,1)
{
  ABORT(n != 1, "Expect exactly one free argument");
  V_Collection *output = V_Collection::make(Value::LIST, m, in_tv);
  if (!output) { return NULL; }
  out_tv[0] = output;
  t = 0;			// Suppress warnings.
  return (QL_Iterator_State*) 1;
}


#define GET_NEEDED_TAG(VAR,NAME) \
VAR = QL_DD::tag_find(NAME); \
if (!VAR) { \
     ql_eval_error(t, "Necessary tag '", NAME, "' not known to database in"); \
     return 0; \
} \

static Material *
get_material(const QL_Term * const t, const Value*in[])
{
  if (in[0]->tv_type() != Value::MATERIAL_POINTER) {
    ql_eval_error(t,
		  "The first argument must be of type 'MATERIAL_POINTER' in");
    return 0;
  }
  return in[0]->down_to_MATERIAL_POINTER();
}

static QL_Iterator_State
*do_primer_case(const QL_Term *t, Value*out[], Value* const primer_v,
		const char *primer_c)
{
  if (primer_v->tv_type() != Value::STRING
      && primer_v->tv_type() != Value::DNA_SEQUENCE) {
    ql_eval_error(t, "Wrong type associted with tag '", primer_c, "' in");
    return 0;
  } else {
    out[0] = primer_v;
    return (QL_Iterator_State*)1;
  }
}


static int get_length(const QL_Term *t, const Step *s,
		      const QL_DD_Tag_Entry *length,
		      const char *length_c, const char *start_c)
{
  Value*length_v = s->find(length->tag_index());
  if (!length_v) {
    ql_eval_error(t, "Must have the tag '", length_c,
		  "' in the same step as tag '", start_c, "' to evaluate");
    return -1;
    
  }
  int ret;
  if (!coerce_to_int(t, "LabBase error: type", length_v, &ret)) return -1;
  return ret;
}



static int find_other_values(const QL_Term *t,
			     Cursor_Step *c,
			     const Step *s,
			     const char *start_c,
			     Sequence **insert_seq)
{
  static QL_DD_Tag_Entry *sequence=0, *insert_start=0, *insert_length;
  if (!sequence) {
    GET_NEEDED_TAG(sequence,"sequence");
    GET_NEEDED_TAG(insert_start,"insert_start");
    GET_NEEDED_TAG(insert_length,"insert_length");
  }
  
  Value *sequence_v=0, *insert_start_v=0, *insert_length_v=0;
  for (; s; s = c->previous()) {
    if (!sequence_v) sequence_v = s->find(sequence->tag_index());
    if (!insert_start_v) {
      insert_start_v = s->find(insert_start->tag_index());
      insert_length_v = s->find(insert_length->tag_index());
    }
    if (sequence_v && !insert_start_v) {
      ql_eval_error(t, "Illegal database contents: ",
		    "a material with 'sequence' and '", start_c, "' tags ",
		    "but no intervening 'insert_start' detected in");
      return 0;
    }
    if ((insert_start_v && !insert_length_v)
	|| (insert_length_v && !insert_start_v)) {
      ql_eval_error(t, "Illegal database contents: ",
		    "step with only one of 'insert_start' or ",
		    "'insert_length' detected in");
      return 0;
    }
    if (insert_start_v && sequence_v) break;
  }
  if (!insert_start_v) {
    ql_eval_error(t, "No 'insert_start' tag found in");
    return 0;
  }
  if (!sequence_v) {
    ql_eval_error(t, "No 'sequence' tag found in");
    return 0;
  }
  int insert_start_int = int(*(Int*)insert_start_v),
  insert_length_int = int(*(Int*)insert_length_v);
  
  if (sequence_v->tv_type() != Value::STRING
      && sequence_v->tv_type() != Value::DNA_SEQUENCE) {
    ql_eval_error(t, "Possible system error in");
    return 0;
  }
  Sequence *seq;
  if (sequence_v->tv_type() == Value::STRING){
    seq = Sequence::make(sequence_v->down_to_STRING());
  } else seq = sequence_v->down_to_DNA_SEQUENCE();
  clear_error();
  String *temp_sv = seq->substring(insert_start_int, insert_length_int);
  if (errmsg()) {
    ql_eval_error(t, errmsg(), " when finding the ``insert'' sequence in");
    return 0;
  }
  ABORT(temp_sv->tv_type() != Value::STRING, "Need a String");
  *insert_seq = Sequence::make(temp_sv);
  return 1;
}


static int 
get_left_tag_entries(const QL_Term * const t,
		     QL_DD_Tag_Entry **primer_p,
		     QL_DD_Tag_Entry **start_p,
		     QL_DD_Tag_Entry **length_p)
{
  static QL_DD_Tag_Entry *primer=0, *start=0, *length=0;
  if (!primer) {
    GET_NEEDED_TAG(primer,"left_primer");
    GET_NEEDED_TAG(start,"left_start");
    GET_NEEDED_TAG(length,"left_length");
  }
  if (start->type()->constructor() != Value::INTEGER) {
    ql_eval_error(t, "Tag 'left_start' must have type 'INTEGER' in");
    return 0;
  }
  if (length->type()->constructor() != Value::INTEGER) {
    ql_eval_error(t, "Tag 'left_length' must have type 'INTEGER' in");
    return 0;
  }
  *primer_p = primer; *start_p = start; *length_p = length;
  return 1;
}


static int 
get_right_tag_entries(const QL_Term * const t,
		      QL_DD_Tag_Entry **primer_p,
		      QL_DD_Tag_Entry **start_p,
		      QL_DD_Tag_Entry **length_p)
{
  static QL_DD_Tag_Entry *primer=0, *start=0, *length=0;
  if (!primer) {
    GET_NEEDED_TAG(primer,"right_primer");
    GET_NEEDED_TAG(start,"right_start");
    GET_NEEDED_TAG(length,"right_length");
  }
  if (start->type()->constructor() != Value::INTEGER) {
    ql_eval_error(t, "Tag 'right_start' must have type 'INTEGER' in");
    return 0;
  }
  if (length->type()->constructor() != Value::INTEGER) {
    ql_eval_error(t, "Tag 'right_length' must have type 'INTEGER' in");
    return 0;
  }
  *primer_p = primer; *start_p = start; *length_p = length;
  return 1;
}



static String *left_final_fn(const QL_Term * const t,
			     const Sequence * const insert,
			     const int start, const int length)
{
  clear_error();
  String *r = insert->substring(start, length);
  if (errmsg()) {
    ql_eval_error(t, errmsg(), " while finding the ``left primer'' ",
		  "sequence in the ``insert'' sequence in");
    return 0;
  }
  return r;
}


static String *right_final_fn(const QL_Term * const t,
			      const Sequence * const insert,
			      const int start, const int length)
{
  char *err1 = " while finding the ``right primer'' sequence in the ";
  char *err2 = "``insert'' sequence in";
  clear_error();
  String *right_primer_reverse = insert->substring(start - length + 1, length);
  if (errmsg()) { ql_eval_error(t, errmsg(), err1, err2); return 0; }
  Sequence *rp_seq = Sequence::make(right_primer_reverse);
  if (errmsg()) { ql_eval_error(t, errmsg(), err1, err2); return 0; }
  String *r = rp_seq->down_to_DNA_SEQUENCE()->reverse_complement();
  if (errmsg()) { ql_eval_error(t, errmsg(), err1, err2); return 0; }
  return r;
}


QL_Iterator_State*
any_primer_seq_pred(const QL_Term *const t, int m, Value* in[],
		    int n, Value *out[],
		    int (*get_tag_entries)(const QL_Term * const t,
					   QL_DD_Tag_Entry **primer_p,
					   QL_DD_Tag_Entry **start_p,
					   QL_DD_Tag_Entry **length_p),
		    
		    String *(*final_fn)(const QL_Term * const t,
					const Sequence * const insert,
					const int start, const int length),
		    const char primer_c[], const char start_c[], const char length_c[])
     
     
{
  abort_m_n(m,1,n,1);
  Material*mtv = get_material(t, in);
  if (!mtv) return 0;
  QL_DD_Tag_Entry *primer=0, *start=0, *length=0;
  if (!get_tag_entries(t,&primer,&start,&length)) return 0;
  Cursor_Step *c = mtv->down_to_MATERIAL_POINTER()->step_cursor();
  if (!c) return 0;
  for (Step *s = c->last(); s; s = c->previous()) {
    Value *primer_v = s->find(primer->tag_index());
    Value *start_v = s->find(start->tag_index());
    if (primer_v && start_v) {
      ql_eval_error(t, "Illegal database contents: a step with both '",
		    primer_c, "' and '", start_c, "' tags, detected in");
      delete c;
      return 0;
    }
    if (primer_v) {
      delete c;
      return do_primer_case(t, out, primer_v, primer_c);
    }
    if (start_v) {
      int length_int = get_length(t, s, length, length_c, start_c);
      if (length_int < 0) { delete c; return 0; }
      int start_int = int(*(Int*)start_v);
      Sequence *insert_seq;
      if (find_other_values(t, c, s, start_c, &insert_seq)) {
	out[0] = final_fn(t, insert_seq,start_int,length_int);
	delete c;
	if (!out[0]) return 0;
	return (QL_Iterator_State*) 1;
      } else { delete c; return 0; }
    }
  }
  delete c;
  return 0;
}


QL_FIRST_HEADER(left_primer_sequence_pred)
{
  return
    any_primer_seq_pred(t, m, in, n, out, get_left_tag_entries, left_final_fn,
			"left_primer", "left_start", "left_length");
}


QL_FIRST_HEADER(right_primer_sequence_pred)
{
  return
    any_primer_seq_pred(t, m, in, n, out, get_right_tag_entries, right_final_fn,
			"right_primer", "right_start", "right_length");
}
#undef GET_NEEDED_TAG


#define TYPE_ERROR  { \
			ql_eval_error(t, "The first argument must have type 'STRING' and the ", \
				      "second argument must have type 'STRING' or 'DNA_SEQUENCE' ", \
				      "for"); \
					return 0; \
					}


#if OBJECTSTORE
#ifndef __REGCMP_H
#include <regcmp.h>
#endif
#else
extern "C" {
  char *re_comp(char *);
  int re_exec(char *); 
}
#endif

QL_FIRST_HEADER(regex_search_pred)
{
  if(out);			// Suppress warnings.  
  if (m != 2) TYPE_ERROR
    if (n > 0) {
      ql_eval_error(t, "Cannot yet bind sub-patterns in");
      return 0;
    }
  if (in[0]->tv_type() != Value::STRING
      || (in[1]->tv_type() != Value::STRING 
	  && in[1]->tv_type() != Value::DNA_SEQUENCE)) TYPE_ERROR
	    String *target_sv;
  if (in[1]->tv_type() == Value::DNA_SEQUENCE) {
    ql_eval_error(t, "Sorry, conversion of 'DNA_SEQUENCE' to ",
		  "'STRING' not yet implemented");
    return 0;
  } else { target_sv = in[1]->down_to_STRING(); }
  char *pattern = in[0]->down_to_STRING()->get();
  char *out_message = 0;
  
  char *comp_result = re_comp(pattern);
  if (comp_result) {
    if (out_message) ql_eval_error(t, out_message,
				   " compiling regular expression '",
				   pattern, "' in");
    else ql_eval_error(t, "Unexplained failure compiling regular expression '",
		       pattern, "' in");
    delete pattern;
    return 0;
  }
  char *temp_pattern = new char [strlen(pattern) + 1];
  safe_strcpy(temp_pattern,pattern);
  // re_exec sometimes re-uses pattern, so we save pattern's
  // contents for error messages.
  char *target = target_sv->get();
  int result = re_exec(target);
  if (result < 0) {
    ql_eval_error(t, "Internal error matching regular expression '",
		  temp_pattern, "' against string '", target, "' in");
  }
  delete pattern;
  delete temp_pattern;
  delete target;
  return (QL_Iterator_State*) (result > 0);
}
#undef TYPE_ERROR


QL_FIRST_HEADER(only_pred)
{
  int total = m + n;
  if (in || out); // Suppress warnings.
  for (QL_Args *a = t->args(); a; a = a->next()) {
    total--;
    if (a->node()) a->node()->set_print();
    else {
      ql_eval_error(t, "All arguments must be variables in");
      return 0;
    }
  }
  ABORT(total, "Wrong number of arguments");
  return (QL_Iterator_State*) 1;
}
