/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: collated_tags.H,v 1.3 1995/02/13 17:49:26 steve Exp $
#ifndef COLLATED_TAGS_H
#define COLLATED_TAGS_H

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef TAG_H
#include "tag.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

class QL_Args;
class QL_Term;
class QL_DD_Tag_Entry;
class Material;
class List_Material;
class Id_Tag_Element;
class Cursor(Id_Tag_Element);
class List(Id_Tag_Element);

class Collated_Tags
// Collated_Tags acts as an intermediary for organizing 
// and checking the tag association arguments
// of insert before creating the associated database objects.
{
private:
  friend class Step; // Step knows how to get the collated tags and values out.

  class ct_entry
    {
    public:
      Value *v[MAX_ARITY * MAX_ARITY];
      // It is convenient to maintain the type information
      // rather than have to go back to the data dictionary.
      // The type information is left behind when
      // the contained Value is copied to the database.
      void increment_count() { _count++; }
      void zero_count() { _count = 0; }
      void set_tag_entry(QL_DD_Tag_Entry *e) { _tag_entry = e; }
      SLOT(QL_DD_Tag_Entry *,tag_entry);
      SLOT(int,count);
    };
  ct_entry _entries[MAX_ARITY];

  int add_arg(QL_Term *,QL_Args *,int);

  ct_entry *find(Tag) const;
  ct_entry *find_or_create(QL_Term *,  QL_DD_Tag_Entry *);
  void add_1_tag_value(ct_entry *, Value *);

  int maybe_do_id_tag(QL_Term*, QL_DD_Tag_Entry*, Value*);  

  List(Id_Tag_Element) *_ids; // All the ids.
  
public:
  int add_tag_value(QL_Term *, QL_DD_Tag_Entry *, Value*);
  int contain(Tag tag) const; // 1 if tag appears; 0 otherwise
  int contain_id_material() const; // 1 if the tags contain any
				    // material such that any
				    // id's should be associated
				    // with that material.

  int value_pointer_number() const; // Return the total number of value pointers;

  String *unique_string_value(Tag) const; // Return a pointer to the single value
						// associated with Tag, cast as a String*.
						// Abort if Tag is not found or does not have
						// exactly one string value.

  Value *find_tag_value(Tag) const;
  
  int unique_boolean_value(Tag) const; // Analogous to unique_string_value.

  int material_number() const; // Return the total number of materials;

  QL_DD_Tag_Entry *contain_create_only_tag() const;
  // Return non-0 iff one of the tags is create_only.

  static Collated_Tags *collate(QL_Term *,QL_Args *, int dd_tags_ok=0);
  // Put all the values for each tag together, and raise user errors
  // under the following conditions:
  // 0. One of the arguments is not a tag association.
  // 1. A tag is data-dictionary and dd_tag_ok==0.
  // 2. The values associated with tag have the wrong type.
  // 3. A single-valued tag occurs more than once.
  // 4. An id tag is already associated (in the database)
  //    with the given value.
  // Other checking has occurred earlier, such checking
  // for the existence of the tag and its arity (which
  // occurs during parsing).

  SLOT(int,tag_number); // Total number of unique tags.
  SLOT(List_Material*,materials); // All the Material_Pointer_Values
				         // associated with tags.

};
#endif
