/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "dd_ops.H"

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_ADORNMENT_H
#include "ql_adornment.H"
#endif

#ifndef QL_ADORNMENT_MAP_H
#include "ql_adornment_map.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef COLLATED_TAGS_H
#include "collated_tags.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef BUILTIN1_H
#include "builtin1.H"
#endif

#ifndef BUILTIN2_H
#include "builtin2.H"
#endif

#ifndef BUILTIN3_H
#include "builtin3.H"
#endif

#ifndef MATERIAL_SET_H
#include "material_set.H"
#endif

#ifndef INSERT_H
#include "insert.H"
#endif

#ifndef DELETE_H
#include "delete.H"
#endif

void DD_Ops::add_hardwired(QL_DD_Entry *e)
{ 
  QL_DD_Entry *e2 = QL_DD::find(e->identifier());
  ABORT(e2,"Attempt to enter duplicate hardwired identifier in data dictionary");
  e->set_next(QL_DD::_dd->_hardwireds); QL_DD::_dd->_hardwireds = e;
}


void DD_Ops::add_hardwired_tags()
{
  add_dd_tag("dd_identifier",Value::STRING, Tag::dd_identifier());
  add_dd_tag("type",Value::TERM, Tag::dd_type());
  add_dd_tag("dd_tag",Value::BOOLEAN, Tag::dd_dd_tag());
  add_dd_tag("id_tag",Value::BOOLEAN, Tag::dd_id_tag());
  add_dd_tag("create_only",Value::BOOLEAN, Tag::create_only());
  add_dd_tag("put_on_history",Value::BOOLEAN, Tag::put_on_history());
  add_dd_tag("segment",Value::INTEGER, Tag::segment());

#ifdef RELEASE2
  add_dd_tag("dd_entry_kind",Value::string_value);
  // It's not possible to create a new dd_entry_kind,
  // so the above will be needed only for data dictionary queries,
  // part of release 2.

  add_dd_tag("dd_implementation",Value::string_value,Value::int_value);
  // We really don't need this one just yet!
#endif

  // Add standard non-data-dictionary tags:
  add_1_hardwired_tag("who", Value::STRING, Tag::who());
  add_1_hardwired_tag("when", Value::DATE, Tag::when());
  add_1_hardwired_tag("step_kind", Value::STRING, Tag::step_kind());
  add_1_hardwired_tag("step_id", Value::INTEGER, Tag::step_id());
  add_1_hardwired_tag("created_material", Value::MATERIAL_POINTER,
		      Tag::created_material(),1);
  add_1_hardwired_tag("material", Value::MATERIAL_POINTER,
		      Tag::material());
}


void DD_Ops::add_1_step_kind_if_needed(char *step_kind)
{
  QL_DD_Entry *f;
  if (!(f = QL_DD::find(step_kind))) {
    QL_DD_Step_Kind_Entry *e
      = PNEW(QL_DD::db(),QL_DD_Step_Kind_Entry,(step_kind,HARDWIRED_USER));
    QL_DD::add_persistent(e);
  } else
    ABORT(f->entry_kind() != QL_DD_Entry::STEP_KIND,
	  "This must be a step kind entry"); 
}


void DD_Ops::add_1_hardwired_tag(char *tag_name, Value::tv_type_id c_type,
				 Tag tag_index, int create_only)
{
  QL_Type_Descriptor td(c_type);
  QL_DD_Tag_Entry *e = new QL_DD_Tag_Entry(tag_name, tag_index,
					   HARDWIRED_USER, &td, create_only);
  add_hardwired(e); 
}


void DD_Ops::add_builtins()
{

  add_builtin(">","bb",gt_pred);
  add_builtin("<","bb",lt_pred);
  add_builtin(">=","bb",ge_pred);
  add_builtin("<=","bb",le_pred);
  add_builtin("<>","bb",ne_pred);
  add_builtin("=", "b",0); // Tag associations.
  add_builtin("==","bb",eq_pred);
  add_builtin("~","bb",regex_search_pred);
  add_builtin("+","*",no_eval_pred);
  add_builtin("-","*",no_eval_pred);
  add_builtin("*","*",no_eval_pred);
  add_builtin("/","*",no_eval_pred);

  add_builtin("all_steps","bf",all_steps_bf_pred_first,all_steps_bf_pred_next);
  // add_builtin("avg_allele_sizes","bf",allele_sizes_pred); NO LONGER USED
  add_builtin("baseline_rusage","",baseline_rusage_pred);
  add_builtin("cardinality","bf",cardinality_pred);
  add_builtin("count","*",count_pred);
  add_builtin("commit","",commit_pred);
  add_builtin("db_size","ff",db_size_pred);
  add_builtin("delete","b",delete_pred);
  add_builtin("delete_entire_material_set","b",delete_entire_value_set_pred);
  add_builtin("delete_entire_temp_material_set","b",
	      delete_entire_temp_value_set_pred);
  add_builtin("delete_entire_value_set","b",delete_entire_value_set_pred);
  add_builtin("delete_entire_temp_value_set","b",
	      delete_entire_temp_value_set_pred);
  add_builtin("element","*",element_pred_first, element_pred_next);
  add_builtin("exists","*",exists_pred);
  add_builtin("gather_in_list","*",gather_in_list_pred);
  add_builtin("gather_in_set","*",gather_in_set_pred);
  add_builtin("incremental_rusage","f",incremental_rusage_pred);
  add_builtin("insert","b",insert_pred);
  add_builtin("insist","b",insist_pred_first, insist_pred_next);
  add_builtin("ith","*",ith_pred_first,ith_pred_next);
  add_builtin("is", "*", is_pred);
  add_builtin("left_primer_sequence","bf",left_primer_sequence_pred);
  add_builtin("length","bf",length_pred);
  add_builtin("make_list","*",make_list_pred);
  add_builtin("material_set","bf",v_set_bf_pred_first,
	      v_set_pred_next, v_set_insert, v_set_delete);
  add_builtin("material_set","bb",v_set_bb_pred_first);
  add_builtin("material_set","f",v_set_f_pred_first,v_set_f_pred_next);
  add_builtin("material_set","b",v_set_b_pred);
  add_builtin("next","bbb", next_pred);
  add_builtin("not","*",   not_pred);
  add_builtin("or","bb",  or_pred_first, insist_pred_next /* sic */); 
  add_builtin(";","bb",   or_pred_first, insist_pred_next /* sic */); 
  add_builtin(";;","bb",   or_pred_first, insist_pred_next /* sic */); 
  add_builtin("only","*", only_pred);
  add_builtin("polymorphic","bbbb", polymorphic_pred);
  add_builtin("reversec", "bf", reversec2_pred);
  add_builtin("reversec", "fb",reversec2_pred);
// Not used:   add_builtin("reversec", "bbbf",reversec_bbbf_pred);
  add_builtin("right_primer_sequence","bf",right_primer_sequence_pred);
  add_builtin("strictly_polymorphic","bbbb", strictly_polymorphic_pred);
  add_builtin("substring", "bbbf",substring_bbbf_pred);
  add_builtin("tag_value", "bf", tag_value_bf_pred);

  add_builtin("temp_material_set", "bf",temp_v_set_bf_pred_first,
	      v_set_pred_next, temp_v_set_insert, temp_v_set_delete);
  add_builtin("temp_material_set", "bb",temp_v_set_bb_pred_first);
  add_builtin("temp_material_set","f",temp_v_set_f_pred_first,v_set_f_pred_next);
  add_builtin("temp_material_set","b",temp_v_set_b_pred);

  add_builtin("temp_value_set", "bf",temp_v_set_bf_pred_first,
	      v_set_pred_next, temp_v_set_insert, temp_v_set_delete);
  add_builtin("temp_value_set", "bb",temp_v_set_bb_pred_first);
  add_builtin("temp_value_set","f",temp_v_set_f_pred_first,v_set_f_pred_next);
  add_builtin("temp_value_set","b",temp_v_set_b_pred);

  add_builtin("true","",true_pred);
//NEWPRED
  add_builtin("tracing_on","f",tracing_on_pred);
  add_builtin("tracing_off","f",tracing_off_pred);

  add_builtin("value_set","bf",v_set_bf_pred_first,
	      v_set_pred_next, v_set_insert, v_set_delete);
  add_builtin("value_set","bb",v_set_bb_pred_first);
  add_builtin("value_set","f",v_set_f_pred_first,v_set_f_pred_next);
  add_builtin("value_set","b",v_set_b_pred);



  add_builtin("var","*",var_pred);

#ifdef FINISHED
  add_builtin("update","b",update_pred,0)
  add_builtin("copy","b",copy_pred,0)
#endif
}

// Add_builtin maintains the invariant that a builtin cannot have
// the same identifier as a non-builtin.

void DD_Ops::add_builtin(char *identifier, char *adornment,
			Builtin_First_Fn first, Builtin_Next_Fn next,
			Write_implementation ii, Write_implementation di,
			Write_implementation ui)
{
  QL_DD_Entry *e = QL_DD::find(identifier);
  QL_Adornment a(adornment);
  if (!e) {
    e = new QL_DD_Builtin_Entry(identifier, ii, di, ui);
    add_hardwired(e);
  } else {
    ABORT(ii || di || ui, "Must be set on initial call to add_builtin");
    ABORT(QL_DD_Entry::BUILTIN != e->entry_kind(),
	  "Attempt to enter builtin with name of a non-builtin")
    QL_Builtin_Implementation *i = ((QL_DD_Builtin_Entry*)e)->adornment_map()->find(a);
    ABORT(i, "Attempt to enter duplicate builtin");
  }
  ((QL_DD_Builtin_Entry*)e)
    ->adornment_map()->add(a, new QL_Builtin_Implementation(first,next));
}


void DD_Ops::add_dd_tag(char *identifier, Value::tv_type_id c_type, Tag tag)
{
  QL_DD_Tag_Entry *e = QL_DD_Tag_Entry::new_dd_tag(identifier, c_type, tag);
  add_hardwired(e);
}


static QL_Type_Descriptor *find_in_term(QL_Term *t)
// Find the value associated with the type tag in "t"
// and return it as a QL_Type_Descriptor*.
// Argument "t" is the argument term of an insert.
// Check that there is exactly one type tag
// in the arguments of t.
{
  QL_Args *args = 0;
  QL_Tag_Assoc *found = 0;
  for (QL_Args *p = t->args(); p; p = p->next()) {
    QL_Tag_Assoc *t_a = p->coerce_to_Tag_Assoc();
    if (t_a->tag()->tag_index() == Tag::dd_type()) {
      if (found) { ql_eval_error(t,"Duplicate tag 'type' in"); return 0; }
      found = t_a;
    }
  }
  if (!found) { 
    ql_eval_error(t, "Required tag 'type' is missing from");
    return 0;
  }
  args = found->args();
  int count;
  QL_Type_Descriptor **out = QL_Type_Descriptor::find_in_args(t, args, &count);
  if (out) {
    QL_Type_Descriptor *r = *out;
    delete [] out;
    return r;
  } else return 0;
}


// Try to generalize extraction of values from
// tag associations in insert, delete, and update arguments.
// for the data dictionary.

// Insert a tag into the data dictionary.
static QL_Iterator_State *dd_insert_tag(QL_Term *t)
{
  Collated_Tags *tags = Collated_Tags::collate(t, t->args(), 1);
  if (!tags) return 0;

  if (!tags->contain(Tag::dd_identifier())) {
    ql_eval_error(t, "Expect dd_identifier=<string> as an argument to");
    return 0;
  }

  int dd_id_tag=0, create_only=0, put_on_history=1, associate_with_id=0,
      segment=-1;
  if (tags->contain(Tag::dd_id_tag())) {
    dd_id_tag = tags->unique_boolean_value(Tag::dd_id_tag());
  }
  if (tags->contain(Tag::create_only())) {
    create_only = tags->unique_boolean_value(Tag::create_only()); 
  }
  if (tags->contain(Tag::put_on_history())) {
    put_on_history = tags->unique_boolean_value(Tag::put_on_history());
  }
  if (tags->contain(Tag::segment())) {
    Value *int_tv = tags->find_tag_value(Tag::segment());
    segment = int(int_tv->down_to_INTEGER());
  }
  if (tags->contain(Tag::associate_with_id()))
    associate_with_id
      = int(*tags->find_tag_value(Tag::associate_with_id())->down_to_INTEGER());

  QL_Type_Descriptor *type = find_in_term(t);
  if (!type) return 0;

  String *dd_identifier = tags->unique_string_value(Tag::dd_identifier());

  char *tag_id = dd_identifier->get();

  if (QL_DD::find(tag_id)) {
    ql_eval_error(t, "'", tag_id, "' is already in the data dictionary in");
    delete type; delete tag_id;
    return 0;
  }

  QL_DD_Tag_Entry *e
    = PNEW(QL_DD::db(),QL_DD_Tag_Entry,(tag_id,
					QL_DD::get_and_increment_tag(),
					"Not implemented",
					type,
                                        create_only,
                                        dd_id_tag,
                                        segment,
                                        put_on_history,
					associate_with_id));
  delete type; delete tag_id;
  QL_DD::add_persistent(e);
  return (QL_Iterator_State*) 1;
}


static int contains_dd_type (QL_Args *args)
// Return 0 iff args contains no tag association with tag
// dd_type.  Abort if any arg is not a tag association.
{
  for ( ; args; args = args->next()) {
    QL_Tag_Assoc *t_a = args->coerce_to_Tag_Assoc();
    ABORT(!t_a, "Only tag associations allowed");
    if (t_a->tag()->tag_index() == Tag::dd_type()) return 1;
  }
  return 0;
}


static char *read_only_dd_identifer(QL_Term *t)
// Read a set of tag/value pairs with only a single
// tag, "dd_identifier".
{
  Collated_Tags *tags = Collated_Tags::collate(t, t->args(), 1);
  if (!tags) return 0;
  if (tags->tag_number() != 1 || tags->value_pointer_number() != 1
      || contains_dd_type(t->args())
      || !tags->contain(Tag::dd_identifier())) {
    ql_eval_error(t, "Expect dd_identifier=<string> as the unique argument to");
    return 0;
  }
  String *dd_identifier = tags->unique_string_value(Tag::dd_identifier());
  char *kind_name = dd_identifier->get();

  if (QL_DD::find(kind_name)) {
    ql_eval_error(t,"'", kind_name, "' is already in the data dictionary in");
    delete kind_name;
    return 0;
  }

  return kind_name;  //dd_identifier->get();
}


static QL_Iterator_State *dd_insert_material_kind(QL_Term *t)
{
  char *kind_name = read_only_dd_identifer(t);
  if (!kind_name) return 0;

//----

  char *primary_id_tag_name = QL_DD_Tag_Entry::primary_id_name(kind_name);
  if (!primary_id_tag_name) { delete kind_name; return 0; }

  if (QL_DD::find(primary_id_tag_name)) {
    ql_eval_error(t,"'", primary_id_tag_name, "' is already in the data dictionary in");
    delete kind_name;
    delete primary_id_tag_name;
    return 0;
  }
  QL_Type_Descriptor td(Value::STRING);
  QL_DD_Tag_Entry *tag_entry
    = PNEW(QL_DD::db(),
      QL_DD_Tag_Entry,(primary_id_tag_name, QL_DD::get_and_increment_tag(),
		      "Not implemented", &td, 1, 1));
  delete primary_id_tag_name;
  ABORT(!tag_entry->id_tag(),"This should have been an id tag");
  QL_DD::add_persistent(tag_entry);
//----

  QL_DD_Material_Kind_Entry *e
    = PNEW(QL_DD::db(),QL_DD_Material_Kind_Entry,(kind_name, "Not implemented",
						   tag_entry->id_index()));
  delete kind_name;

  QL_DD::add_persistent(e);
  return (QL_Iterator_State*) 1;
}


static QL_Iterator_State *dd_insert_step_kind(QL_Term *t)
{
  char *kind_name = read_only_dd_identifer(t);
  if (!kind_name) return 0;

  QL_DD_Step_Kind_Entry *e
    = PNEW(QL_DD::db(),QL_DD_Step_Kind_Entry,(kind_name, "Not implemented"));
  delete kind_name;
  QL_DD::add_persistent(e);
  return (QL_Iterator_State*) 1;
}


void DD_Ops::add_dd_kind(char *identifier, Write_implementation ii)
{
  QL_DD_DD_Kind_Entry *e
    = new QL_DD_DD_Kind_Entry(identifier, ii);
  add_hardwired(e);
}


void DD_Ops::add_all_dd_kinds()
{
  add_dd_kind("tag",dd_insert_tag);
  add_dd_kind("builtin",0);
  add_dd_kind("data_dictionary_kind",0);
  add_dd_kind("dd_step_kind",dd_insert_step_kind);
  add_dd_kind("material_kind",dd_insert_material_kind);
  add_dd_kind("type_constructor",0);
}


void DD_Ops::add_type_constructors()
{
  add_type_constructor("STRING");
  add_type_constructor("INTEGER");
  add_type_constructor("FLOAT");
  add_type_constructor("DATE");
  add_type_constructor("BOOLEAN");
  add_type_constructor("SET",1);
  add_type_constructor("LIST",1);
  // add_type_constructor("BAG",1);
  add_type_constructor("MATERIAL_POINTER");
  add_type_constructor("DNA_SEQUENCE");
  add_type_constructor("STEP_POINTER");
  add_type_constructor("TERM");
  add_type_constructor("SLIST",1);
  add_type_constructor("TUPLE",1);
}


void DD_Ops::add_type_constructor(char* id, int has_args)
{
  Value::tv_type_id type = Value::char2type(id);
  ABORT(Value::UNDEFINED == type, "Illegal type string");
  QL_DD_Type_Constructor_Entry *e
    = new QL_DD_Type_Constructor_Entry(id, type, has_args);
  add_hardwired(e);
}


void DD_Ops::add_hardwireds()
{
  add_builtins();
  add_all_dd_kinds();
  add_type_constructors();
  add_hardwired_tags();
  add_1_step_kind_if_needed("create");
}
