/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef ID_INDEX_ENTRY_H
#define ID_INDEX_ENTRY_H

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef LIST_H
#include "list.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif


// if we are not compiling a main memory version
#if defined(MAIN_MEMORY)
#define USE_TEXAS_INDEX 1
#else
#define USE_TEXAS_INDEX (defined(TEXAS) && defined(USE_INDEX))
#endif

class Id_Index_Entry {
public:
#if USE_TEXAS_INDEX
  Id_Index_Entry(Material* m=0):_material(m){};
#else
    // This block is included when Texas Btree index are not used
    // in which case char* id data memeber holds the string key

  Id_Index_Entry(Material *m,char*);

  ~Id_Index_Entry();

#ifdef OBJECTSTORE
  os_backptr b;
#endif

  SLOT(char *,id); // A pointer into the String associated
		    // with an id tag in some step.
		    // For char * data members we 
		    // must employ ``user-controlled index maintenance''
		    // ---See ObjectStore User Guide page 168.
		    // This caveat probably applies only if the
		    // the contents of the char * are modified
		    // without modifying its address, but
		    // I have not confirmed this.
#endif

  SLOT(Material *,material);  // The material with which the id
			       // is associated.
};


class Str_Ptr
{
  SLOT(String*,s);
public:
  int operator< (const Str_Ptr sp1) const { return (*s() < *sp1.s()); }
  int operator> (const Str_Ptr sp1) const { return (*s() > *sp1.s()); }
  int operator== (const Str_Ptr sp1) const { return (*s() == *sp1.s()); }
  Str_Ptr(String* s1=0):_s(s1){};
};


#if defined(OBJECTSTORE)
#define ID_INDEX_SET Set<Id_Index_Entry*>
#define ID_INDEX_CURSOR os_Cursor<Id_Index_Entry*>
#define ID_INDEX_CURSOR_OF() os_cursor(os_cursor::unsafe)
#elif USE_TEXAS_INDEX
#ifndef BTREE_H
#include "btree.hh"
#endif
#define ID_INDEX_SET BTree<Str_Ptr,Id_Index_Entry>
#define ID_INDEX_CURSOR BTCursor<Str_Ptr,Id_Index_Entry>
#define ID_INDEX_CURSOR_OF() cursor_of()
#else
Listdeclare(Id_Index_Entry);
#define ID_INDEX_SET List(Id_Index_Entry)
#define ID_INDEX_CURSOR Cursor(Id_Index_Entry)
#define ID_INDEX_CURSOR_OF() cursor_of()
#endif

#endif
