/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: insert.C,v 1.5 1995/03/01 18:11:00 steve Exp $
#ifndef IOSTREAMH
#include <iostream.h>
#endif

#ifndef FSTREAMH
#include <fstream.h>
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef COLLATED_TAGS_H
#include "collated_tags.H"
#endif

#ifndef STEP_H
#include "step.H"
#endif

#ifndef ID_INDEX_ENTRY_H
#include "id_index_entry.H"
#endif

#ifndef INSERT_H
#include "insert.H"
#endif


class ID_INDEX_SET;

class QL_Iterator_State;
typedef QL_Iterator_State *(*Insert_implementation)(QL_Term *);

ofstream *log_stream = 0;

static QL_Iterator_State *insert_builtin_or_dd(QL_Term *t)
{
  QL_DD_Entry *entry = t->dd_entry();
  Insert_implementation i = entry->implementation_of_insert();
  if (!i) ql_eval_error(t, "No way to insert into");
  else return i(t);
}


static QL_Iterator_State *insert_material(QL_Term *t)
{
  // Insert an instance of a material.
  // We already know that the material kind exists, or
  // we wouldn't have gotten this far.  So all
  // we have to do is create the material object
  // and create its ``create'' step.

  Collated_Tags *tags = Collated_Tags::collate(t, t->args());
  if (!tags) return 0;

  // We don't want to put anything in the database
  // Until we know that everything is o.k.

  QL_DD_Material_Kind_Entry *material_kind_entry
    = t->dd_entry()->down_to_MATERIAL_KIND();
  // t is the term for the predicate <material>(....).
  // t->dd_entry() is the data dictionary entry for
  //  <material>, the material kind.

  char *primary_id_tag
    = QL_DD_Tag_Entry::primary_id_name(material_kind_entry->identifier());

  // Now we check the tags:
  QL_DD_Tag_Entry *id_entry = QL_DD::tag_find(primary_id_tag);

  ABORT(!id_entry, "The primary identifier for the material is unknown");

  Tag id = id_entry->tag_index();

#ifdef OBJECTSTORE
#define GET_TAG_INDEX(TAG) \
  QL_DD_Tag_Entry * CAT_ID2(TAG,_entry); \
  Tag TAG = 0; \
  if (!TAG) CAT_ID2(TAG,_entry) = QL_DD::tag_find("TAG"); \
  ABORT(!CAT_ID2(TAG,_entry),"We must have the tag 'TAG'"); \
  TAG = CAT_ID2(TAG,_entry)->tag_index();

#else
#define GET_TAG_INDEX(TAG) \
  QL_DD_Tag_Entry * CAT_ID2(TAG,_entry); \
  Tag TAG = 0; \
  if (!TAG) CAT_ID2(TAG,_entry) = QL_DD::tag_find(#TAG); \
  ABORT(!CAT_ID2(TAG,_entry),"We must have the tag 'TAG'"); \
  TAG = CAT_ID2(TAG,_entry)->tag_index();
#endif

  GET_TAG_INDEX(step_kind);
  GET_TAG_INDEX(created_material);
#undef GET_TAG_INDEX

  if (!tags->contain(id)) {
    ql_eval_error(t, "Tag '", primary_id_tag, "' required in");
    delete primary_id_tag;
    return 0;
  }
  delete primary_id_tag;

  if (!tags->contain(Tag::who())) {
    ql_eval_error(t, "Tag 'who' required in");
    return 0;
  }

  if (!tags->contain(Tag::when())) {
    ql_eval_error(t, "Tag 'when' required in");
    return 0;
  }

  if (tags->contain(Tag::created_material())) {
    ql_eval_error(t, "User must not supply the tag 'created_material' in");
    return 0;
  }

  if (tags->contain(Tag::step_kind())) {
    ql_eval_error(t, "User must not supply the tag 'step_kind' in");
    return 0;
  }

  // Make sure that the id doesn't already exist.
  ID_INDEX_SET* main_index = id_entry->id_index();
  ABORT(!main_index, "There should have been an id index");

  char *identifier_value = tags->unique_string_value(id)->get();
  if (!id_entry->check_duplicate(t,identifier_value)) {
    delete identifier_value;
    return 0;
  }

  if (tags->contain_id_material()) {
    ql_eval_error(t, "No material tag should be supplied when inserting ",
		  "a material in");
    delete identifier_value;
    return 0;
  }

  // Up till now, nothing has gone into the database.

  Material *m
   = PNEW(QL_DD::material_segment(),
          Material,(material_kind_entry,identifier_value));
  delete identifier_value;

  if (!tags->add_tag_value(t,created_material_entry,m)) {
    ABORT(1,"Transaction backout not implemented");
    return 0;
  }

  // String *kind_value = String::make(material_kind_entry->identifier());

  // if (!tags->add_tag_value(t,step_kind_entry,kind_value)) {
  //  ABORT(1,"Transaction backout not implemented");
  //  return 0;
  // }

  // Abstract the follwing as a static belonging to Step:

  QL_DD_Step_Kind_Entry *s_k_entry
    = QL_DD::find("create")->down_to_STEP_KIND();
  clear_error();

  Step *the_create_step  = PNEW(QL_DD::step_segment(),Step,(tags,s_k_entry));
  if (errmsg()) { 
    delete the_create_step; ql_eval_error(t, errmsg(), "in");
    ABORT(1,"Transaction backout not implemented");
    return 0; }

  return (QL_Iterator_State *) 1;
}


static QL_Iterator_State *insert_step(QL_Term *t)
{
  Collated_Tags *tags = Collated_Tags::collate(t, t->args());
  if (!tags) return 0;

  // We don't want to put anything in the database
  // Until we know that everything is o.k.

  QL_DD_Step_Kind_Entry *step_kind_entry = t->dd_entry()->down_to_STEP_KIND();
  // t is the term for the predicate <step_kind>(....).
  // t->dd_entry() is the data dictionary entry for<step_kind>.
  // We should not have gotten this far unless t's entry is a step
  // kind entry.


  if (!strcmp(step_kind_entry->identifier(),"create")) {
    ql_eval_error(t, "Cannot insert a second 'create' step in");
    return 0;
  }
  if (tags->contain_create_only_tag()) {
    ql_eval_error(t, "Cannot use the create_only tag '",
		  tags->contain_create_only_tag()->identifier(), 
                  "' in a step of kind '",
		  step_kind_entry->identifier(), "' in");
    return 0;
  }

  if (!tags->contain(Tag::who())) {
    ql_eval_error(t, "Tag 'who' required in");
    return 0;
  }
  
  {
    Value *date = tags->find_tag_value(Tag::when());
    if (!date) {
      ql_eval_error(t, "Tag 'when' required in");
      return 0;
    }
    time_t time = date->down_to_DATE()->time_in_seconds();

    {
      Cursor_Material *c = tags->materials()->cursor_of();
      for (Material *mv = c->first(); mv; mv=c->next()) {
	if (mv->creation_timestamp() > time) {
	  ql_eval_error(t, "Inserted step is earlier than the first step for ",
			"material with id ", mv->id(),
			" in");
	  delete c;
	  return 0;
	}
      }
      delete c;
    }
  }

  // Check that there is at least one material tag: 
  if (tags->material_number() < 1) {
    ql_eval_error(t, "Need at least one material in");
    return 0;
  }

  clear_error();
  Step *s = PNEW(QL_DD::step_segment(),Step,(tags, step_kind_entry));
  if (errmsg()) { delete s; ql_eval_error(t, errmsg(), "in"); return 0; }
  

  return (QL_Iterator_State *) 1;
}


static Value::tv_type_id insert_types[1] = { Value::TERM };
BUILTIN_FIRST_DEFINITION(insert_pred,1,0,insert_types,1,0)
{
  QL_Term *arg_t = in_tv[0]->down_to_TERM();

  if (out_tv) t = 0; m = n = 0; // Supress compiler warnings.
  // Switch on the type of the data-dictionary entry:
  QL_DD_Entry::kind k = arg_t->dd_entry()->entry_kind();
  switch (k) {
  case QL_DD_Entry::TAG:
    ql_eval_error(arg_t, "Cannot perform an insert on tag");
    break;
  case QL_DD_Entry::BUILTIN: return insert_builtin_or_dd(arg_t);
  case QL_DD_Entry::MATERIAL_KIND: return insert_material(arg_t);
  case QL_DD_Entry::STEP_KIND: return insert_step(arg_t);
  case QL_DD_Entry::DATA_DICTIONARY_KIND: return insert_builtin_or_dd(arg_t);
  default: ABORT(1, "Illegal data-dictionary entry type");
    break;
  }
}
