/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include <string.h>

#include "ql_adornment.H"

QL_Adornment::QL_Adornment(const char *p)
{
  _len = strlen(p);
  if ('*' == *p) { 
    ABORT(_len > 1, "Illegal `*' adornment");
    _any = 1; *_value = '\0';
  }
  else {
    ABORT(_len >= MAX_ARITY, "Adornment exceeds maximum arity");
    _any = 0;
    for (unsigned i = 0; i < _len; i++) {
      _value[i] = p[i];
      ABORT((_value[i] != 'b' && _value[i] != 'f'),
            "Illegal character in adornment");
    }
    _value[_len] = '\0';
  }
}


int QL_Adornment::operator == (const QL_Adornment& a) const
{
  return (a._any && _any) || (a._len == _len && !strcmp(a._value,_value));
}


QL_Adornment::operator char*()
{
  if (_any) return "*";
  return _value;
}


int QL_Adornment::distance(const QL_Adornment &a) const
{
  if (a._any && _any) return -1;
  if (a._len != _len) return -1;
  int r = 0;
  for (int i = 0; i < _len; i++)
    if (a._value[i] != _value[i]) {
      if (_value[i] == 'b') r++;
      else return -1;
    }
  return r;
}


void QL_Adornment::add(char c)
{
  ABORT(('f'!= c && 'b' != c), "Illegal character added to adornment");
  ABORT(_any,"Cannot add to an ``any'' adornment");
  _len++;
  ABORT(_len >= MAX_ARITY, "Adornment exceeds maximum arity");
  _value[_len - 1] = c;
  _value[_len] = '\0';
}


char *QL_Adornment::braces() const
{
  static char tmp[MAX_ARITY + 3];
  *tmp = '[';
  strcpy(tmp +1, _value);
  *(tmp + _len + 1) = ']';
  *(tmp + _len + 2) = '\0';
  return tmp;
}


int QL_Adornment::number_free() const
{
  int sum = 0;
  for (int i=0; i<_len; i++) if ('f' ==_value[i]) sum++;
  return sum;
}


int QL_Adornment::number_bound() const
{
  return _len - this->number_free();
}


char QL_Adornment::operator [] (int i) const
{
  ABORT(i >= _len, "Adornment index out of bounds");
  return _value[i]; 
}
