/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef QL_ADORNMENT_MAP_H
#define QL_ADORNMENT_MAP_H

// An Adornment_Map is part of the data dictionary.
// An adornment map maps adornments to predicate implementations
// (as represented by function entry points).

// In the context of a particular predicate identifier, an adornment map
// tells how to evaluate that predicate for a particular adornment.

// Adornment maps are relevant only for builtin predicates.
// Consequently, they need never be persistent.

// The implementation is essentially an a-list because
// we think the number of possible adornments will be small.
// We break out the adornment map because predicates
// derived from tags or kinds dispatch on adornments
// differently.

class QL_Builtin_Implementation;

#ifndef QL_ADORNMENT_H
#include "ql_adornment.H"
#endif

class QL_Adornment_Map
{
private:
class QL_Adornment_Node
  {
 public:
    QL_Adornment_Node
      (QL_Adornment &adornment, QL_Builtin_Implementation *implementation) :
    _next(NULL), _adornment(adornment), _implementation(implementation) {}

    SLOT(QL_Adornment_Node*,next);
    SLOT(QL_Adornment,adornment);
 public:
    void set_next(QL_Adornment_Node *x) { _next = x; }

    static QL_Builtin_Implementation *lookup(QL_Adornment_Node *N, QL_Adornment &a)
      {
	QL_Adornment any("*");
	while (N) {
	  if (a == N->_adornment || N->_adornment == any) return N->_implementation;
	  N = N->_next;
	}
	return NULL;
      }

 protected:
    QL_Builtin_Implementation *_implementation;
  };

  QL_Adornment_Node *_entries;

public:
  QL_Builtin_Implementation *find(QL_Adornment &adornment)
    { return QL_Adornment_Node::lookup(_entries,adornment);  } 

  QL_Adornment_Map() : _entries(0) {}

  int add(QL_Adornment &adornment, QL_Builtin_Implementation *implementation)
    {
      if (QL_Adornment_Node::lookup(_entries,adornment)) return 0;
      else {
	QL_Adornment_Node *i = new QL_Adornment_Node(adornment, implementation);
	i->set_next(_entries);
	_entries = i;
	return 1;
      }
    }

};
#endif
