/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_args.C,v 1.3 1995/02/13 17:50:20 steve Exp $
#include "ql_args.H"


extern int respond_with_control_A;


void QL_Bindings::print()
{ 
  QL_Binding_Node *p;
  int obey_print = 0;
  if (_nodes) {
    for (p = _nodes; p; p=p->_next) if (p->_print) { obey_print = 1; break; }
    char comma = '\0';
    for (p = _nodes; p; p=p->_next) {
      if (p->_print || !obey_print) {
	if (comma) cout << comma;
	comma = respond_with_control_A ? '\034' : ',';
	char equal = respond_with_control_A ? '\034' : '=';
	if (p->value())
	  cout << p->_identifier << equal << *p->value();
	else
	  cout << p->_identifier << equal << '_';
      }
    }
    cout << "\n";
  }
}


QL_Binding_Node *QL_Bindings::add(String *string)
{
  char *identifier = string->get();
  if (!_nodes) {
    _nodes = new QL_Binding_Node(identifier); // NewC
    delete identifier;
    return _nodes;
  }
  else {
    for (QL_Binding_Node *p = _nodes; 1; p = p->_next) {
      if (!strcmp(identifier, p->_identifier)) { delete identifier; return p; }
      if (!p->_next) {
	p->_next = new QL_Binding_Node(identifier);
	delete identifier;
	return p->_next;
      }
    }
  }
}


QL_Args::QL_Args(Value *tv, QL_Args *next) :
  _arg(tv), _node(0), _next(next), _garbage_next(_garbage_list)
    { _garbage_list = this; }


QL_Args::QL_Args(QL_Binding_Node *node, QL_Args *next) :
  _arg(0), _node(node), _next(next), _garbage_next(_garbage_list)
    { _garbage_list = this; }


QL_Args *QL_Args::set_next(QL_Args *x)
{
  _next = x;
  for (QL_Args *p = this; p; p = p->next())
    ABORT(p->node() && p->arg(), "Illegal arg");
  return this;
}


QL_Tag_Assoc *QL_Args::coerce_to_Tag_Assoc()
{
  return
    (arg() 
     && arg()->tv_type() == Value::TERM 
     && ((QL_Term*)arg())->is_tag_assoc())
    ? (QL_Tag_Assoc*) arg() : 0;
}


QL_Adornment QL_Args::adornment()
{
  QL_Adornment a;
  for (QL_Args *p = this; p; p = p->next())
    a.add(p->_node && ! p->_node->value() ? 'f': 'b');
  return a;
}


QL_Args *QL_Args::_garbage_list = 0;
void QL_Args::gc() {
  while (_garbage_list) {
    QL_Args *p = _garbage_list->_garbage_next;
    delete _garbage_list;
    _garbage_list = p;
  }
}
