/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_args.H,v 1.3 1995/02/13 17:50:22 steve Exp $
#ifndef QL_ARGS_H
#define QL_ARGS_H

#ifndef IOSTREAMH
#include "iostream.h"
#endif

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef QL_ADORNMENT_H
#include "ql_adornment.H"
#endif

#ifndef QL_GC_H
#include "ql_gc.H"
#endif

#ifndef __STRING_H
#include <string.h>
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

class QL_Args;

class QL_Binding_Node
{
public:
  QL_Binding_Node(char *identifier) : _value(0), _next(0), _print(0)
    {
      _identifier = new char[strlen(identifier) + 1]; // NewC
      strcpy(_identifier,identifier);
    }

  void add_value(Value *v) { _value = v; }
  void set_print() { _print = 1; };
  SLOT(Value*,value);
  SLOT(char*,identifier);   // The name of the variable.
  SLOT(int,print);          // If 0, print only if all nodes have _print == 0.
private:
  friend class QL_Bindings;
  ~QL_Binding_Node()
    {
      delete [] _identifier;
    }
  QL_Binding_Node *_next;
};


class QL_Bindings
{
public:
  QL_Bindings() : _nodes(0) {}

  ~QL_Bindings()
    {
      QL_Binding_Node *q, *p = _nodes;
      while (p) { q = p->_next; delete p; p = q; }
    }

  QL_Binding_Node *add(String *string);
  // Return the binding node for the variable with name string.
  // Add a node if necessary.

  void print();

private:
  QL_Binding_Node *_nodes;

};


class QL_Args {
public:
  // Rely on the value of node() to determine whether a given
  // argument is a Value or a pointer to a binding node.

  QL_Args(Value *, QL_Args*next=0);

  QL_Args(QL_Binding_Node *, QL_Args *next=0);

  QL_Args *set_next(QL_Args *x);

  QL_Adornment QL_Args::adornment();

  Value *value() const {
    if (_node) return _node->value();
    else return _arg;
  }

  QL_Tag_Assoc *coerce_to_Tag_Assoc();

  int arity() const {
    const QL_Args *p = this;
    int r = 0;
    while (p) { r++; p = p->_next; }
    return r;
  }

  SLOT(QL_Binding_Node*,node);
  SLOT(Value*,arg);
  SLOT(QL_Args*,next);

private:
  QL_GC_MEMBERS(QL_Args);
};


#endif


