/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_builtin_impl.C,v 1.3 1995/02/13 17:50:24 steve Exp $
#ifndef QL_BUILTIN_IMPL_H
#include "ql_builtin_impl.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

QL_Term *tv2term(QL_Term *t, Value *tv)
{
  if (!tv) return 0;
  if (tv->tv_type() == Value::TERM) return tv->down_to_TERM();
  if (tv->tv_type() == Value::STRING) {
    char *id = tv->down_to_STRING()->get();
    QL_DD_Entry *e = QL_DD::find(id);
    delete id;
    if (e) return new QL_Term(e,0,t->parse_position());
  }
  return 0;
}


int
type_ok(QL_Term *t, Value **v, Value::tv_type_id type)
{
  return ((type == (*v)->tv_type()) 
	  // Coerce a string to a term, if possible.
	  || (type == Value::TERM && (*v = tv2term(t,*v))));
}

// static int type_ok(QL_Term *t, Value **v, Value::tv_type_id type)
// {
//   Value::tv_type_id vtype = (*v)->tv_type();
//   if (Value::INTEGER == vtype) vtype = Value::INTEGER32;
//   return ((type == v_type) 
// 	  // Coerce a string to a term, if possible.
// 	  || (type == Value::TERM && (*v = tv2term(t,*v))));
// }

int check_arg_type(QL_Term *t,
		   Value **v, Value::tv_type_id type, int arg_number)
{
  if (type_ok(t, v, type)) return 1;
  char arg_buf[100];
  sprintf(arg_buf, "%d", arg_number+1);
  if (type == Value::TERM)
    ql_eval_error(t, "Argument ", arg_buf, " must be a term in");
  else
    ql_eval_error(t, "Argument ", arg_buf, " must be a '",
		  Value::type2char(type), "' in");
  return 0;
}


QL_Iterator_State *
inject_and_typecheck_first(QL_Term*t, int m, Value*in[], int n, Value*out[],
                           Value::tv_type_id in_types[], int type_number,
			   int expected_m, int expected_n, int,
                           BUILTIN_FIRST_HEADER((*first_fn)))
{
  ABORT(type_number < 1 && expected_m != 0,
	"Please supply at least one type");
  char x_buf[100];
  if (expected_m > 0 && expected_m != m) {
    sprintf(x_buf, "%d", expected_m);
    ql_eval_error(t, "Expected ", x_buf, " bound arguments for");
    return 0;
  }
  if (expected_n > 0 && expected_n != n) {
    sprintf(x_buf, "%d", expected_n);
    ql_eval_error(t, "Expected ", x_buf, " free arguments for");
    return 0;
  }
  for (int i=0; i < m; i++) {
    Value::tv_type_id this_type
      = i < type_number ? in_types[i] : in_types[type_number-1];
    ABORT(!in[i],"Must have non-0 value");
    if (this_type != Value::UNDEFINED && !check_arg_type(t,&in[i],this_type,i))
      return 0;
  }
  QL_Iterator_State *r = first_fn(t, m, in, n, out);
  if (r) for (int j=0; j<n; j++) {
    ABORT(!out[j],"Must have a non-0 Value");
  }
  return r;
}


QL_Iterator_State *
inject_next(QL_Term*t, QL_Iterator_State *s,
	    int n, Value*out[], int,
	    BUILTIN_NEXT_HEADER((*next_fn)))
{
  QL_Iterator_State *r = next_fn(t, s, n, out); 
  if (r) for (int j=0; j<n; j++) {
    ABORT(!out[j],"Must have a non-0 Value");
  }
  return r;
}

int
coerce_to_int(const QL_Term *t,  const char *msg, Value* v, int*out)
{
  ABORT(!out,"Need to supply second argument");
  if (v->tv_type() == Value::INTEGER) {
      *out = int(*v->down_to_INTEGER());
      return 1;
    } else if (v->tv_type() == Value::INTEGER32) {
      *out = int(*v->down_to_INTEGER32());
      return 1;
    } else {
      ql_eval_error(t, msg, " must be an 'INTEGER' or 'INTEGER32' in");
      return 0;
    }
}

