/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


// $Id: ql_eval.C,v 1.6 1995/02/13 17:50:42 steve Exp $
#include "ql_eval.H"

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_ITERATOR_H
#include "ql_iterator.H"
#endif

#ifndef QL_ERROR_H
#include "ql_error.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef TAG_QUERY_H
#include "tag_query.H"
#endif

#ifndef STEP_QUERY_H
#include "step_query.H"
#endif

int inside_exists = 0;


static int ql_eval1(QL_Term*, QL_Bindings*);


// Top level eval function, must be called only once per query if the
// user-level (nested) commits and aborts are enabled.
void QL_eval(QL_Term *term, QL_Bindings *bindings)
{
  if (ql_eval) {
    // The functions QL_DD::begin_nested, etc, were an (unsuccessful) attempt
    // do get finer-grained durability in ObjectStore.
    // QL_DD::begin_nested();
    int r = ql_eval1(term, bindings);
    if (ql_eval) {
      // QL_DD::commit_nested();
      if (r) cout << "YES\n";
      else cout << "NO\n";
    } else 
      /* QL_DD::abort_nested() */;
  }
}


static int ql_eval1(QL_Term *term, QL_Bindings *bindings)
{
  int r = 0;
  ABORT(!ql_eval,"Should not eval when ql_eval is 0");
  if (!term) { bindings->print(); return 1; }
  else {
    QL_Iterator *iterator = create_iterator(term);
    if (iterator) {
      Value::pushmem();
      while (iterator->true_bindings() && ql_eval)
	r |= ql_eval1(term->next(),bindings);
      Value::popmem();
      delete iterator;
    }
  }
  return r;
}


int QL_eval_no_print(QL_Term *term)
{
  int r = 0;
  ABORT(!ql_eval,"Should not eval when ql_eval is 0");
  ABORT(!term,"Term must be non-zero");
  QL_Iterator *iterator = create_iterator(term);
  if (iterator) {
    while (iterator->true_bindings() && ql_eval)
      if (term->next())	r += QL_eval_no_print(term->next());
      else r++;
    delete iterator;
  }
  return r;
}


int QL_eval_exists(QL_Term *term)
{
  ABORT(!ql_eval,"Should not eval when ql_eval is 0");
  ABORT(!term,"Term must be non-zero");

  // This 'inside_exists' variable is hack to prevent step queries from
  // executing inside 'exists' predicates.  We really need an additional
  // (optional) function for each predicate, which the iterator can call to
  // destroy any state associated with the evaluation of a predicate.  In
  // addition, QL_eval_exists shouldn't destroy iterators that yield true
  // bindings.  Instead, it should gather them (recursively), to be destroyed
  // only when the 'exits' iterator is destoryed.  Otherwise we loose the
  // bindings.

  int old_inside_exists = inside_exists;
  inside_exists = 1;

  QL_Iterator *iterator = create_iterator(term);
  if (iterator) {
    while (iterator->true_bindings() && ql_eval)
      if (!term->next()) { 
	inside_exists = old_inside_exists; delete iterator; return 1;
      }
      else {
	if (QL_eval_exists(term->next())) {
	  inside_exists = old_inside_exists; delete iterator; return 1;
	}
      }
  }
  inside_exists = old_inside_exists; 
  delete iterator;
  return 0;
}


void *
QL_eval_with_aggregate(QL_Term *error_term,
                       QL_Term *term,
		       void *running_value,
		       QL_Binding_Node *bindings_to_use[],
                       int binding_number,
		       void*(*update_value)(QL_Term *, void*, Value**, int))
{
  ABORT(!ql_eval,"Should not eval when ql_eval is 0");
  ABORT(!term,"Term must be non-zero");
  ABORT(!error_term, "Error_term must be non-zero");
  QL_Iterator *iterator = create_iterator(term);
  if (iterator) {
    while (iterator->true_bindings() && ql_eval) {
      if (term->next())	QL_eval_with_aggregate(error_term, term->next(),
					       running_value, bindings_to_use, 
					       binding_number, update_value);
      else {
	Value **values = new Value*[binding_number];
        for (int i=0; i < binding_number; i++) {
	  Value *ba = bindings_to_use[i]->value();
	  if (!ba) {
	    ql_eval_error(error_term, "Variable ",
			  bindings_to_use[i]->identifier(),
			  " is unbound in");
	    delete [] values;
	    delete iterator;
	    return 0;
	  }
	  values[i] = ba;
	}
	running_value
	  = update_value(error_term, running_value, values, binding_number);
	delete [] values;
	if (! running_value) {
	  delete iterator;
	  return 0;
	}
      }
    }
    delete iterator;
  }
  return running_value;
}

// Probably not used. 
//static void cannot_evaluate(QL_Term *t)
//{
//  ql_eval_error(t, "Evaluation of adornment",
//		t->adornment().braces(), " not implemented");
//}


static QL_Iterator *create_material_kind_iterator(QL_Term *t)
{
  if (t->number_free() + t->number_bound() != 1) {
    ql_eval_error(t, "System supports only a unary version of marker kind");
    return 0;
  }
  if (t->number_free() == 1) {
    QL_Iterator *i
      = new QL_Iterator(t, 0, 1, Material::iterator_first_0_1,
			    Material::iterator_next_0_1);
    return i;
  }
  else {
    return new QL_Iterator(t, 1, 0, Material::iterator_first_1_0, 0);
  }
}


static QL_Iterator *create_data_dictionary_kind_iterator(QL_Term *t)
{
  ql_eval_error(t, "Queries not supported on data_dictionary_kind");
  return 0;
}


static QL_Iterator *create_builtin_iterator(QL_Term *t)
{
  if (!t->set_implementation()) return 0;
  
  QL_FIRST_HEADER((*first))
    =  (Builtin_First_Fn) t->implementation()->implementation_of_first();
  QL_NEXT_HEADER((*next))
    = (Builtin_Next_Fn)	t->implementation()->implementation_of_next();
    
  return new QL_Iterator(t, t->number_bound(), t->number_free(), first, next);
}


QL_Iterator *create_iterator(QL_Term *t)
     // The binding nodes are available as QL_Args in term.
{
  t->set_adornment_info();	// The adornment info depends on the binding
  // pattern, which cannot be determined until
  // evaluation time.
  // The name of this function
  
  QL_DD_Entry::kind k = t->dd_entry()->entry_kind();
  switch (k) {
  case QL_DD_Entry::MATERIAL_KIND: return create_material_kind_iterator(t);
  // create_material_kind_query either needs to set the implementation
  // of the term, or it needs to have a new constructor
  // for QL_Iterators.  The second alternative seems cleaner.
				   
  case QL_DD_Entry::TAG: return Tag_Query::create_iterator(t);
				   
  case QL_DD_Entry::STEP_KIND: return create_step_kind_iterator(t);
				   
  case QL_DD_Entry::DATA_DICTIONARY_KIND: return create_data_dictionary_kind_iterator(t);
				   
  case QL_DD_Entry::BUILTIN: return create_builtin_iterator(t);
				   
  case QL_DD_Entry::TYPE_CONSTRUCTOR:
    ql_eval_error(t,"A type name is not legal as");
    return 0;

 default: ABORT(1,"Unkown dd_entry kind");
 }
}
