/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#include "step.H"

#ifndef BASICS_H
#include "basics.H"
#endif

#ifndef VALUE_H
#include "value.H"
#endif

#ifndef COLLATED_TAGS_H
#include "collated_tags.H"
#endif

#ifndef QL_DD_ENTRY_H
#include "ql_dd_entry.H"
#endif

#ifndef MATERIAL_H
#include "material.H"
#endif

#ifndef ID_INDEX_ENTRY_H
#include "id_index_entry.H"
#endif

#ifndef QL_ARGS_H
#include "ql_args.H"
#endif

#ifndef QL_TERM_H
#include "ql_term.H"
#endif

#ifndef QL_DD_H
#include "ql_dd.H"
#endif

#ifndef ID_TAG_ELEMENT_H
#include "id_tag_element.H"
#endif


Step::Step(Collated_Tags *tags,QL_DD_Step_Kind_Entry*s_k_e)
: Value(STEP_POINTER), _delete_marked(0), _tag_array(0), _value(0), _step_kind(s_k_e),
  _timestamp(0),
  _tag_number(tags->tag_number() - 1) // We treat 'when' specially.
{
  P_SEGMENT *def_seg = QL_DD::default_segment();

// This is for tag clustering only
#ifdef OBJECTSTORE
  _tag_array = PNEW_ARRAY(QL_DD::tag_array_segment(),Tag,tag_number());
#else
  //_tag_array = Tag_Store::array(tag_number()); // for tag clustering only
#endif

  _tag_array = PNEW_ARRAY(QL_DD::tag_array_segment(),Tag,tag_number());
  _value = PNEW_ARRAY(def_seg,Value*,tag_number());

  Value **target = &_value[0];
  Tag *target_tag = &_tag_array[0];

  for (int i = 0; i < tags->tag_number(); i++) {
    // i is the index into tags->_entries.
    
    if (tags->_entries[i].tag_entry()->tag_index() == Tag::when())
      _timestamp = tags->_entries[i].v[0]->down_to_DATE()->time_in_seconds();
    else {
      // A lot of the following code is vestigial, from the
      // time when we had n-valued tags and tags with
      // "arity" != 1.
      *target_tag = tags->_entries[i].tag_entry()->tag_index();
      Value *tv = tags->_entries[i].v[0];
      ABORT(tv->tv_type() == Value::STEP_POINTER,
	    "Steps cannot be associated with tags");
      if (tv->tv_type() == Value::MATERIAL_POINTER)
	*target = tv->down_to_MATERIAL_POINTER();
      else *target = tv->save(def_seg);
      ABORT(!*target,"Storing null value");
      target++; target_tag++;
    }
  }
  ABORT(tag_number() != (target - &_value[0]),"Storage error creating Step");
  ABORT(tag_number() != (target_tag - &_tag_array[0]),"Storage error");
  ABORT(timestamp()==0, "step must have a when tag");
  add_to_materials(tags);
}


Date Step::when() const
{
  return *find(Tag::when())->down_to_DATE();
}


Value *Step::find(Tag tag) const
{
  if (tag == Tag::when()) return Date::make(timestamp());
  if (tag == Tag::step_kind()) return String::make(step_kind()->identifier());
  for(unsigned int i = 0; i < tag_number(); i++)
    if (_tag_array[i] == tag) return _value[i];
  return 0;
}


int Step::contains_materials(List_Value* value_set) const
// We simply check each Value* in the "this"
// to see if it is one of the Value *s in value_set.
// If all the Value *s in value_set are in "this",
// then return 1.
{
  Cursor_Value *c = value_set->cursor_of();
  for (Value *v = c->first(); v; v = c->next()) {
    for(unsigned int i = 0; i < tag_number(); i++) {
      if (_value[i] == v) // These are Materials, so pointer equality is appropriate.
	break;
    }
    if (i >= tag_number()) { delete c; return 0; }
  }
  delete c; return 1;
}


void Step::add_to_materials(Collated_Tags *tags)
{
  Cursor_Material *m_cursor = tags->_materials->cursor_of();
  for (Material* m_p = m_cursor->first() ; m_p; m_p = m_cursor->next()) {
    m_p->add_step(this);
    Cursor_Id_Tag_Element *t_cursor = tags->_ids->cursor_of();
    if (1) {  // Put code here to see if the material is on a
              // tag with ``associated_with_id''.
      for (Id_Tag_Element* it = t_cursor->first()
	   ; it; it = t_cursor->next())	{
        ID_INDEX_SET  * index = it->tag_entry()->id_index();
	Value *tv_id = it->tv();
	ABORT(tv_id->tv_type() != Value::STRING, "Only string ids allowed");
#if USE_TEXAS_INDEX
	Value *tmp = find(it->tag_entry()->tag_index());
        // We need to find the value associated with the tag in `this', because
        // we need the persistent copy of the String.  The copy in tags->_ids is
        // probably transient.
        ABORT(!tmp,"Cannot find tag index in `this'");
        Str_Ptr sp(tmp->down_to_STRING());
        Id_Index_Entry index_entry(m_p);
// Note that the transient object key sp and value index_entry are actualy 
// copied in their _persistent_ counterparts in the Btree node (key, value) 
        index->insert(sp,index_entry);
#else
	char *char_id = tv_id->down_to_STRING()->get();
	index->insert(PNEW(QL_DD::db(),Id_Index_Entry,(m_p,char_id)));
	delete char_id;
#endif
      }
      delete t_cursor;
    }
  }
  delete m_cursor;
}


void Step::print(ostream& os) const
{
  Date dt = when();
  os << "step(" << step_kind()->identifier()
    << "," << "(when=" << dt << "))";
}


int Step::compare(const Value& a) const
{
  if (a.tv_type() != Value::STEP_POINTER) {
    set_error("Attempt to compare a 'STEP_POINTER' to a non 'STEP_POINTER'");
    return -1;
  }
  else return (((Step*)&a) != this);
}


Step* Step::save(P_SEGMENT *)
{
  ABORT(1, "Do not call save on Step's--must already be in db");
  return 0;
}
