/* 
 * Copyright (c) 1995
 *       Whitehead Institute for Biomedical Research. All rights reserved.
 *
 * Released under terms specified in the Copyright file present in the top
 * directory of this distribution.
 */


#ifndef TAG_H
#define TAG_H

#include <limits.h>

#ifndef LIST_H
#include "list.H"
#endif


// Class Tag is essentially a wraper for type short, and is used as the
// abbreviated representation of tags in the database.  (An alternative
// representation would be a pointer to the data dictionary entry for the
// tag.)
class Tag 
{
  short _value;
  
public:
  Tag(int i) : _value(i) {}
  Tag() : _value(SHRT_MIN) {} // A junk value, need for array intialization.
  operator ==(const Tag t) const { return t._value == _value; }
  operator !() { return _value == 0; }

  // The following functions return Tags for built-in tags.
  static Tag dd_identifier() { return Tag(-1); }
  static Tag dd_type() { return Tag(-2); }
  static Tag dd_dd_tag() { return Tag(-3); }
  static Tag dd_id_tag() { return Tag(-4); }
  static Tag create_only() { return Tag(-5); }
  static Tag segment() { return Tag(-6); }
  static Tag when() { return Tag(-7); }
  static Tag who() { return Tag(-8); }
  static Tag step_kind() { return Tag(-9); }
  static Tag step_id() { return Tag(-10); }
  static Tag created_material() { return Tag(-11); }
  static Tag material() { return Tag(-12); }
  static Tag put_on_history() { return Tag(-13); }
  static Tag associate_with_id() { return Tag(-14); }
};

Listdeclare(Tag);

// #if 0
#ifndef OBJECTSTORE
class Tag_Store
{
   const int initial_count = 1 << 13;
   int _count_remaining;
   Tag *_storage;
   Tag_Store() : 
     _count_remaining(initial_count),
     _storage(PNEW_ARRAY(,Tag,initial_count)) {}
public:
   static Tag* array(int count) {
       static Tag_Store *current_tag_store= NULL;
       if(!current_tag_store || current_tag_store->_count_remaining < count)
          current_tag_store = PNEW(,Tag_Store,);
       Tag* r = current_tag_store->_storage;
       current_tag_store->_storage += count;
       current_tag_store->_count_remaining -= count;
       return r; }
};
#endif
//#endif

#endif

